/*
 * 
 * The JANClient is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 *
 * The JANClient is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License.
 * If not, see <http://www.gnu.org/licenses/>. 
 * 
 * The goal for this small project was a way to give, in order to use Astrometry.Net 
 * Web Portal in Java projects without to install any Python software.
 * 
 * The java client for astrometry.net
 * To read the FITS please use FV software, for example
 * http://heasarc.gsfc.nasa.gov/docs/software/ftools/fv/
 * 
 * the content type was changed to: "application / x-www-form-urlencoded"
 * 
 * after the version 938 will be following included:
 *1 If astrometry is a failure or that the time is exceeded (in any case astrometry is not a success), 
 *  it is possible to have a ./download/file_name.failure.txt file.
 *2 if You have a bad internet connection, it is possible to receive only the file_name.calibration file ( no file_name.fits files).
 *
 *3 //TODO not realised yet :To improve the computation time, it is possible to transmit, scale_lower; scale_upper; scale_est; center_ra; center_dec; radius.?
 *  http://astrometry.net/doc/net/api.html
 */
package astrometry.net.client.java;

import com.jeamy.jascom.Telescope;
import com.jeamy.jascom.helper.Chooser;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.Charset;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Pattern;
import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.mime.HttpMultipartMode;
import org.apache.http.entity.mime.MultipartEntity;
import org.apache.http.entity.mime.content.FileBody;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.ContainerFactory;
import org.json.simple.parser.JSONParser;

/**
 *
 * @author Dr. Valentin Steinhauer (Telescope Engineering Company - TEC - USA)
 * @eMail tec.de@t-online.de
 * @see http://www.telescopengineering.com
 * @version 24.02.2014
 */
public class JANClient {

    final static String version = "$Rev: 949 $";
    final static Logger log = Logger.getLogger(JANClient.class);
    static boolean nofits = false;
    static String url = "http://nova.astrometry.net";
    static String apikey = "xxxxxxxxxx";//please set Your private apikey in the config.properties 
    static String session = "";
    static String fileName = "test_jpg.jpg";//default value
    static String subid = "";
    static String[] jobids = {""};
    static String[] jobstaties = {""};
    static String[] user_images = {""};
    static String userid = "";
    static String delimiter = ";";
    static Long waitingTime = 10000L; //10000 msc default
    static int howManyTimes = 10;//10 default
    static boolean debug = false;
    static String out = "./";
    static boolean ascom = false;
    static double racenter;
    static double deccenter;
    DecimalFormat dfRF = new DecimalFormat("#.##");
    DecimalFormat dfRF3 = new DecimalFormat("#.###");
    static StringBuilder outputCSV = new StringBuilder();// to append the results from tar file

    public JANClient() {
        PropertyConfigurator.configure("./log4j.properties");
        setConfig();
    }

    /**
     * @param args the command line arguments You must input only the name of
     * image file in jpg or png or bmp or gif or tif format, for example: java
     * -jar JANClient.jar IMG4551.jpg It contains following results in
     * directory: IMG4551.ANNOTATED_FULL.jpg - with all identified objects
     * IMG4551.AXY.fits - fits file with pixel positions IMG4551.RDLS.fits -
     * fits with RA,DEC coordinates IMG4551.WCS.fits IMG4551.csv - the file with
     * callibration data for the image,IMG4551.NEW-IMAGE.fits - the image with
     * callibration (for example for aladin) and annotations file In this
     * version we have the possibility to work with list of files from tar
     * archive.
     */
    public static void main(String[] args) {
        System.out.println(args[0]);
        JANClient client = new JANClient();
        try {
            File f = new File(args[0]);
            boolean b = f.isFile();
            if (!b) {
                client.janclient_exit();
            }
        } catch (Exception e) {
            client.janclient_exit();
        }
        fileName = args[0];
        List list = new ArrayList();
        if (fileName.indexOf(".tar") > 0) {
            list = untar(fileName, out);
        }
        System.out.print("Java Astrometry.net Client " + version + " is running .");
        log.info("---------------- next task ------------------------------");
        log.info("Java Astrometry.net Client " + version + " is running for :" + fileName);
        if (list.isEmpty()) {
            boolean b = newSession(fileName);// b is not used yet
            System.exit(0);
        } else {
            Iterator it = list.iterator();
            outputCSV.append(",,Image center,\nImage,job id,RA,DEC\n");
            while (it.hasNext()) {
                fileName = (String) it.next();
                boolean b = newSession(out + fileName);// b is not used yet
            }
            try (FileOutputStream out1 = new FileOutputStream(out + "output.csv"); BufferedOutputStream bufferedOutStream = new BufferedOutputStream(out1)) {
                byte[] bytes = outputCSV.toString().getBytes();
                bufferedOutStream.write(bytes);
                bufferedOutStream.flush();
                out1.close();
            } catch (Exception e) {
                System.out.println(e.toString());
                log.error(e);
            }
            System.exit(0);
        }
    }

    private static boolean newSession(String localFileName) {
        JANClient client = new JANClient();
        try {
            session = client.loginPost();
            System.out.print(".");
            if (!session.isEmpty() && !session.equals("None")) {
                subid = client.postFile(session, localFileName);
            } else {
                System.out.println("The session is empty for the apikey = " + apikey + ".\n");
                log.error("The session is empty for the apikey = " + apikey);
                System.exit(0);
            }
            System.out.print(".");
            // workflow
            if (!subid.isEmpty() && !subid.equals("")) {
                boolean b = false;
                for (int i = 0; i < howManyTimes; i++) {
                    Thread.sleep(waitingTime);
                    jobids = client.getJobids(subid);
                    jobstaties = client.getJobStaties();
                    System.out.print(".");
                    b = client.testResult();
                    if (b) {
                        break;
                    }
                }
                if (!b) {
                    System.out.print("Attention ! : The job results are not reached.\n The run must be repeated, sorry. \n");
                    log.error("The job results are not reached. The run must be repeated for file:" + localFileName);
                    System.exit(0);
                }
                if (debug) {
                    System.out.print("\nSubmission ID:" + subid + "\n");
                }
                for (int i = 0; i < jobids.length; i++) {
                    if (jobstaties[i].equals("success")) {
                        if (!nofits) {
                            client.getFile(jobids[i], new File(out + client.replaceSuffix(fileName, "WCS.fits")), "/wcs_file/");
                            System.out.print(".");
                            client.getFile(jobids[i], new File(out + client.replaceSuffix(fileName, "ANNOTATED_FULL.jpg")), "/annotated_full/");
                            System.out.print(".");
                            client.getFile(jobids[i], new File(out + client.replaceSuffix(fileName, "AXY.fits")), "/axy_file/");
                            System.out.print(".");
                            client.getFile(jobids[i], new File(out + client.replaceSuffix(fileName, "NEW-IMAGE.fits")), "/new_fits_file/");
                        }
                        System.out.print(".");
                        client.getAnnotations(jobids[i]);
                        System.out.print(".");
                        client.getCalibration(jobids[i]);
                        if (ascom) {
                            client.synchronization(racenter, deccenter);
                        }
                    }
                }
            }
        } catch (Exception e) {
            System.out.println("main:" + e);
            log.error(e);
            return false;
        }
        return true;
    }

    private static List untar(String localFileName, String localOut) {
        List pictures = new ArrayList();
        try {
            try (TarArchiveInputStream myTarFile = new TarArchiveInputStream(new FileInputStream(new File(localFileName)))) {
                TarArchiveEntry entry = null;
                String individualFiles;
                int offset;
                FileOutputStream outputFile = null;
                while ((entry = myTarFile.getNextTarEntry()) != null) {
                    individualFiles = entry.getName();
                    byte[] content = new byte[(int) entry.getSize()];
                    offset = 0;
                    System.out.println("File Name in TAR File is: " + individualFiles);
                    log.info("File Name in TAR File is: " + individualFiles);
                    pictures.add(individualFiles);
//                    System.out.println("Size of the File is: " + entry.getSize());
//                    System.out.println("Byte Array length: " + content.length);                 
                    myTarFile.read(content, offset, content.length - offset);
                    outputFile = new FileOutputStream(new File(localOut + individualFiles));
                    IOUtils.write(content, outputFile);
                    outputFile.close();
                }
            }
        } catch (Exception ep) {
            System.out.println(ep.toString());
            log.error(ep);
        }
        return pictures;
    }

    private void setConfig() {
        try {
            Properties prop = new Properties();
            FileInputStream stream = new FileInputStream("./config.properties");
            prop.load(stream);
            apikey = prop.getProperty("apikey");
            waitingTime = Long.parseLong(prop.getProperty("waitingTime"));
            howManyTimes = Integer.parseInt(prop.getProperty("howManyTimes"));
            debug = Boolean.parseBoolean(prop.getProperty("debug"));
            out = prop.getProperty("out");
            ascom = Boolean.parseBoolean(prop.getProperty("ascom"));
            nofits = Boolean.parseBoolean(prop.getProperty("nofits"));
            try {
                File f = new File(out);
                if (!f.isDirectory()) {
                    System.out.println("The directory out = " + out + " does not exist. ");
                    log.error("The directory out = " + out + " does not exist. ");
                    System.exit(0);
                }
            } catch (Exception eout) {
                System.err.println(eout);
                log.error(eout);
                System.exit(0);
            }
            int n = out.lastIndexOf("/");
            if (n != out.length() - 1) {
                out = out + "/";
            }
        } catch (IOException | NumberFormatException e) {
            System.out.println(e.getMessage());
            log.error(e);
        }
    }

    String replaceSuffix(String fileName, String suffix) {
        int lastPointPosition = fileName.lastIndexOf(".");
        if (lastPointPosition < 0) {
            System.out.println("The name of image file is NOT correct:" + fileName);
            log.error("The name of image file is NOT correct:" + fileName);
            System.exit(0);
        }
        StringBuilder sb = new StringBuilder();
        sb.append(fileName.substring(0, lastPointPosition + 1));
        sb.append(suffix);
        return sb.toString();
    }

    static JSONObject getLoginJSON(String s) {
        JSONObject obj = new JSONObject();
        obj.put("apikey", s);
        return obj;
    }

    String loginPost() {
        StringBuilder sb = new StringBuilder();
        try {
            URL u = new URL(url + "/api/login");
            URLConnection uc = u.openConnection();
            uc.setDoOutput(true);
            uc.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
            try (PrintWriter pw = new PrintWriter(uc.getOutputStream())) {
                JSONObject jsonObject1 = getLoginJSON(apikey);
                pw.println("request-json=" + jsonObject1.toJSONString());
            }
            //StringBuilder sb = new StringBuilder();
            try (BufferedReader in = new BufferedReader(new InputStreamReader(uc.getInputStream()))) {
                String res = in.readLine();
                sb.append(res);
                if (debug) {
                    System.out.println(res);
                }
            }
        } catch (Exception e) {
            System.out.println("Login:" + e);
            log.error("login:" + e);
        }

//        HttpClient httpClient = new DefaultHttpClient();
//        HttpPost postRequest = new HttpPost(url + "/api/login");
//        JSONObject jsonObject = getLoginJSON(apikey);
//        StringEntity input = new StringEntity("request-json=" + jsonObject.toJSONString());
//        input.setContentType("application/x-www-form-urlencoded");
//        postRequest.setEntity(input);
//        HttpResponse response = httpClient.execute(postRequest);
//        if (response.getStatusLine().getStatusCode() != 200) {
//            throw new RuntimeException("Failed : HTTP error code : "
//                    + response.getStatusLine().getStatusCode());
//        }
//        BufferedReader br = new BufferedReader(
//                new InputStreamReader((response.getEntity().getContent())));
//        String output;
//        StringBuilder sb = new StringBuilder();
//        if (debug) {
//            System.out.println("Output from Server ....");
//        }
//        while ((output = br.readLine()) != null) {
//            if (debug) {
//                System.out.println(output);
//            }
//            sb.append(output);
//        }
//        httpClient.getConnectionManager().shutdown();
        return jsonGetValue(sb.toString(), "session");
    }

    String jsonGetValue(String sjson, String key) {
        String value = new String();
        try {
            JSONParser parser = new JSONParser();
            ContainerFactory containerFactory = new ContainerFactory() {
                @Override
                public List creatArrayContainer() {
                    return new LinkedList();
                }

                @Override
                public Map createObjectContainer() {
                    return new LinkedHashMap();
                }
            };
            Map json = (Map) parser.parse(sjson, containerFactory);
            Iterator iter = json.entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry entry = (Map.Entry) iter.next();
                if (entry.getKey().toString().toLowerCase().equals(key)) {
                    value = entry.getValue().toString();
                }
            }
        } catch (Exception e) {
            if (e.toString().equals("Unexpected character (<) at position 0.")) {
                log.warn("jsonGetValue:" + e);
            } else {
                System.out.println(e);
                log.error("jsonGetValue:" + e);
            }
        }
        if (debug) {
            System.out.println("get value through key:" + key + " value=" + value);//test
        }
        return value;
    }

    String postFile(String session, String fileName) throws Exception {
        int GOOD_RETURN_CODE = 200;
        File imageFile = new File(fileName);
        HttpClient httpclient = new DefaultHttpClient();
        HttpPost httppost;
        MultipartEntity reqEntity;
        JSONObject jsonObject = getUploadJSON(session);
        httppost = new HttpPost(url + "/api/upload/");
        reqEntity = new MultipartEntity(
                HttpMultipartMode.BROWSER_COMPATIBLE);
        reqEntity.addPart("request-json", new StringBody(jsonObject.toJSONString(), "application/json", Charset.defaultCharset()));
        reqEntity.addPart("file", new FileBody(imageFile, fileName, "application/octet-stream", Charset.defaultCharset().toString()));
        httppost.setEntity(reqEntity);
        String result = "";
        try {
            HttpResponse response = httpclient.execute(httppost);
            StatusLine statusLine = response.getStatusLine();
            int statusCode = statusLine.getStatusCode();
            StringBuilder builder = new StringBuilder();
            if (statusCode == GOOD_RETURN_CODE) {
                HttpEntity entity = response.getEntity();
                try (InputStream content = entity.getContent()) {
                    BufferedReader reader = new BufferedReader(new InputStreamReader(content));
                    String line;
                    while ((line = reader.readLine()) != null) {
                        builder.append(line).append("\n");
                    }
                } catch (Exception esymbol) {
                    System.out.println("postFile esymbol:" + esymbol.getMessage());
                }
                result = builder.toString();
            } else {
                System.out.println(statusCode);
                log.info("satusCode:" + statusCode);
            }
        } catch (IOException | IllegalStateException e) {
            System.out.println("postFile:" + e.getMessage());
            log.error("postFile:" + e);
        }
        httpclient.getConnectionManager().shutdown();
        String success = jsonGetValue(result, "status");
        String subidLocal = jsonGetValue(result, "subid");
        String hash = jsonGetValue(result, "hash");
        if (debug) {
            System.out.println("upload status:" + success);
            System.out.println("upload subid:" + subidLocal);
            System.out.println("upload hash:" + hash);
        }
        return subidLocal;
    }

    static JSONObject getUploadJSON(String session) {
        JSONObject obj = new JSONObject();
        obj.put("publicly_visible", "y");
        obj.put("allow_modifications", "d");
        obj.put("session", session);
        obj.put("allow_commercial_use", "d");
        return obj;
    }

    boolean testResult() {
        boolean b = false;
        for (int i = 0; i < jobids.length; i++) {
            if (!jobids[i].isEmpty() && jobstaties[i].equals("success")) {
                b = true;
                break;
            }
        }
        return b;
    }

    String[] getJobids(String subid) {
        String[] jobidsLocal;
        StringBuilder builder = new StringBuilder();
        String result = "";
        try {
            HttpClient httpclient = new DefaultHttpClient();
            HttpGet httpget = new HttpGet(url + "/api/submissions/" + subid);
            HttpResponse response = httpclient.execute(httpget);
            if (debug) {
                System.out.println("getJobids gets: " + response.getStatusLine());
            }
            HttpEntity entity = response.getEntity();
            try (InputStream content = entity.getContent()) {
                BufferedReader reader = new BufferedReader(new InputStreamReader(content));
                String line;
                while ((line = reader.readLine()) != null) {
                    builder.append(line).append("\n");
                }
            } catch (Exception esymbol) {
                System.out.println("getJobids symbol:" + esymbol.getMessage());
                log.error("getJobids symbol:" + esymbol);
            }
            result = builder.toString();
        } catch (IOException | IllegalStateException e) {
            System.out.println("getJobids:" + e.getMessage());
            log.error("getJobids:" + e.getMessage());
        }
        String jobs = jsonGetValue(result, "jobs");
        jobs = jobs.replaceAll(Pattern.quote("["), "");
        jobs = jobs.replaceAll(Pattern.quote("]"), "");
        if (debug) {
            System.out.println("jobs line:" + jobs);
        }
        jobidsLocal = jobs.split(",");
        if (debug) {
            if (jobidsLocal.length >= 1) {
                System.out.println("JOB IDs:");
                for (int i = 0; i < jobidsLocal.length; i++) {
                    System.out.println(jobidsLocal[i]);
                }
            }
        }
        userid = jsonGetValue(result, "user");
        if (debug) {
            System.out.print("\n Userid:" + userid + "\n");
        }
        String user_imagesLocal = jsonGetValue(result, "user_images");
        user_imagesLocal = user_imagesLocal.replaceAll(Pattern.quote("["), "");
        user_imagesLocal = user_imagesLocal.replaceAll(Pattern.quote("]"), "");
        if (debug) {
            System.out.println(" user_images line:" + user_imagesLocal);
        }
        user_images = user_imagesLocal.split(",");
        if (debug) {
            if (user_images.length >= 1) {
                System.out.println("\nuser_images:\n");
                for (int i = 0; i < user_images.length; i++) {
                    System.out.println(user_images[i]);
                }
            }
        }
        return jobidsLocal;
    }

    String[] getJobStaties() {
        String[] staties = new String[jobids.length];
        for (int i = 0; i < jobids.length; i++) {
            StringBuilder builder = new StringBuilder();
            String result = "";
            try {
                HttpClient httpclient = new DefaultHttpClient();
                HttpGet httpget = new HttpGet(url + "/api/jobs/" + jobids[i]);
                HttpResponse response = httpclient.execute(httpget);
                if (debug) {
                    System.out.println("getJobSaties gets: " + response.getStatusLine());
                }
                HttpEntity entity = response.getEntity();
                try (InputStream content = entity.getContent()) {
                    BufferedReader reader = new BufferedReader(new InputStreamReader(content));
                    String line;
                    while ((line = reader.readLine()) != null) {
                        builder.append(line).append("\n");
                    }
                } catch (Exception esymbol) {
                    System.out.println("getJobStaties symbol:" + esymbol.getMessage());
                    log.error("getJobStaties symbol:" + esymbol);
                }
                result = builder.toString();
            } catch (IOException | IllegalStateException e) {
                System.out.println("getJobStaties:" + e.getMessage());
                log.error("getJobStaties:" + e.getMessage());
            }
            String status = jsonGetValue(result, "status");
            if (debug) {
                if (!jobids[i].isEmpty()) {
                    System.out.println(jobids[i] + " " + status);
                }
            }
            staties[i] = status;
        }
        return staties;
    }

    void getCalibration(String jobid) {
        StringBuilder builder = new StringBuilder();
        String result = "";
        try {
            HttpClient httpclient = new DefaultHttpClient();
            HttpGet httpget = new HttpGet(url + "/api/jobs/" + jobid + "/calibration");
            HttpResponse response = httpclient.execute(httpget);
            if (debug) {
                System.out.println("getCalibration gets: " + response.getStatusLine());
            }
            HttpEntity entity = response.getEntity();
            try (InputStream content = entity.getContent()) {
                BufferedReader reader = new BufferedReader(new InputStreamReader(content));
                String line;
                while ((line = reader.readLine()) != null) {
                    builder.append(line).append("\n");
                }
            } catch (Exception esymbol) {
                System.out.println("getCalibration esymbol:" + esymbol.getMessage());
                log.error("getCalibration esymbol:" + esymbol);
            }
            result = builder.toString();
        } catch (IOException | IllegalStateException e) {
            System.out.println("getCalibration:" + e.getMessage());
            log.error("getCalibration:" + e.getMessage());
        }
        String parity = jsonGetValue(result, "parity");
        String orientation = jsonGetValue(result, "orientation");
        String pixscale = jsonGetValue(result, "pixscale");
        String radius = jsonGetValue(result, "radius");
        String radecimal = jsonGetValue(result, "ra");
        String decdecimal = jsonGetValue(result, "dec");
        racenter = Double.parseDouble(radecimal);
        deccenter = Double.parseDouble(decdecimal);
        String sracenter = new String(dfRF3.format(Double.parseDouble(radecimal)));
        sracenter = sracenter.replaceAll(",", ".");
        String sdecdecimal = new String(dfRF3.format(Double.parseDouble(decdecimal)));
        sdecdecimal = sdecdecimal.replaceAll(",", ".");
        outputCSV.append(fileName).append(",").append(jobid).append(",")
                .append(sracenter).append(",").append(sdecdecimal).append("\n");
        if (debug) {
            System.out.println("Job " + jobid);
        }
        System.out.println("\nImage " + fileName + "\nCalibration of the image:");
        //System.out.println("     Parity      " + parity);

        System.out.println("     Orientation: Up is " + dfRF.format(Double.parseDouble(orientation)) + " degrees E of N");
        System.out.println("     Pixscale    " + dfRF.format(Double.parseDouble(pixscale)) + "(arcsec/pixel)");
        System.out.println("     Radius      " + dfRF3.format(Double.parseDouble(radius)) + "(deg)");
        System.out.println("     Center RA(deg/hms)     " + dfRF3.format(Double.parseDouble(radecimal)) + "(deg) " + deg2HMS(Double.parseDouble(radecimal)));
        System.out.println("     Center DEC(deg/hms)    " + dfRF3.format(Double.parseDouble(decdecimal)) + "(deg) " + deg2HMS(Double.parseDouble(decdecimal)));
        StringBuilder sb = new StringBuilder();

        log.info("Image " + fileName);
        log.info("     Orientation: Up is " + dfRF.format(Double.parseDouble(orientation)) + " degrees E of N");
        log.info("     Pixscale    " + dfRF.format(Double.parseDouble(pixscale)) + "(arcsec/pixel)");
        log.info("     Radius      " + dfRF3.format(Double.parseDouble(radius)) + "(deg)");
        log.info("     Center RA(deg/hms)     " + dfRF3.format(Double.parseDouble(radecimal)) + "(deg) " + deg2HMS(Double.parseDouble(radecimal)));
        log.info("     Center DEC(deg/hms)    " + dfRF3.format(Double.parseDouble(decdecimal)) + "(deg) " + deg2HMS(Double.parseDouble(decdecimal)));

        /**
         * ******* line 1 in csv *******
         */
        sb.append("parity");
        sb.append(delimiter);
        sb.append("orientation(deg)");
        sb.append(delimiter);
        sb.append("pixscale(arsec/pix)");
        sb.append(delimiter);
        sb.append("radius(deg)");
        sb.append(delimiter);
        sb.append("center-ra(deg)");
        sb.append(delimiter);
        sb.append("center-dec(deg)");
        sb.append("\n");
        /**
         * ******* line 2 in csv *********
         */
        sb.append(parity);
        sb.append(delimiter);
        sb.append(dfRF.format(Double.parseDouble(orientation)));
        sb.append(delimiter);
        sb.append(dfRF.format(Double.parseDouble(pixscale)));
        sb.append(delimiter);
        sb.append(dfRF3.format(Double.parseDouble(radius)));
        sb.append(delimiter);
        sb.append(radecimal);
        sb.append(delimiter);
        sb.append(decdecimal);
        String sf = replaceSuffix(out + fileName, "calibration.csv");
        saveText(sf, sb.toString());
    }

    void getAnnotations(String jobid) {
        StringBuilder builder = new StringBuilder();
        String result = "";
        try {
            HttpClient httpclient = new DefaultHttpClient();
            HttpGet httpget = new HttpGet(url + "/api/jobs/" + jobid + "/annotations");
            HttpResponse response = httpclient.execute(httpget);
            if (debug) {
                System.out.println("getAnnotations gets: " + response.getStatusLine());
            }
            HttpEntity entity = response.getEntity();
            try (InputStream content = entity.getContent()) {
                BufferedReader reader = new BufferedReader(new InputStreamReader(content));
                String line;
                while ((line = reader.readLine()) != null) {
                    builder.append(line).append("\n");
                }
            } catch (Exception esymbol) {
                System.out.println("getAnnotations esymbol:" + esymbol.getMessage());
                log.error("getAnnotations esymbol:" + esymbol.getMessage());
            }
            result = builder.toString();
        } catch (IOException | IllegalStateException e) {
            System.out.println("getAnnotations:" + e.getMessage());
            log.error("getAnnotations:" + e.getMessage());
        }

        StringBuilder csvText = new StringBuilder();
        try {
            JSONParser parser = new JSONParser();
            Object obj = parser.parse(result);
            JSONObject jsonObject = (JSONObject) obj;
            JSONArray array = (JSONArray) jsonObject.get("annotations");
            if (debug) {
                System.out.println("annotation size:" + array.size() + "\n");
            }
            Iterator iter = array.iterator();
            while (iter.hasNext()) {
                JSONObject entry = (JSONObject) iter.next();
                //System.out.println(entry.toString());
                double x = (Double) entry.get("pixelx");
                double y = (Double) entry.get("pixely");
                StringBuilder sb = new StringBuilder();
                JSONArray namesArray = (JSONArray) entry.get("names");
                Iterator it = namesArray.iterator();
                while (it.hasNext()) {
                    String ent = (String) it.next();
                    sb.append(ent);
                    sb.append(" ");
                }
                double radius = (Double) entry.get("radius");
                String type = (String) entry.get("type");
                if (debug) {
                    System.out.println(x + " " + y + " " + radius + " " + sb.toString() + " " + type);
                }
                csvText.append(dfRF.format(x));
                csvText.append(delimiter);
                csvText.append(dfRF.format(y));
                csvText.append(delimiter);
                csvText.append('"').append(sb.toString().trim()).append('"');//Es gibt doppelte Namen
                csvText.append(delimiter);
                csvText.append(dfRF3.format(radius));
                csvText.append(delimiter);
                csvText.append(type);
                csvText.append("\n");
            }
            String sf = replaceSuffix(out + fileName, "annotations.csv");
            saveText(sf, csvText.toString());
        } catch (Exception e) {
            System.out.println("annotations:" + e);
            log.error("annotations:" + e);
        }
    }

    /**
     * Download the url content and save into the file. It can be used to
     * download images,pdf,flash,fits etc. It must be reworked to exchange the
     * deprecations
     *
     * external parameter: url
     *
     * @param jobid
     * @param file
     * @param suburl
     * @return
     */
    public boolean getFile(String jobid, File file, String suburl) {
        HttpClient httpclient = new DefaultHttpClient();
        HttpGet httpget = new HttpGet(url + suburl + jobid);
        boolean result = true;
        try {
            HttpResponse response = httpclient.execute(httpget);
            HttpEntity entity = response.getEntity();
            InputStream content = entity.getContent();
            try (FileOutputStream out1 = new FileOutputStream(file)) {
                byte[] buffer = new byte[1024];
                int count;
                while ((count = content.read(buffer)) != -1) {
                    out1.write(buffer, 0, count);
                }
                out1.flush();
            }
        } catch (IOException | IllegalStateException e) {
            System.out.println("getCorrFile:" + e.getMessage());
            log.error("getCorrFile:" + e.getMessage());
            result = false;
        }
        return result;
    }

    void saveText(String filePath, String text) {
        try {
            File textFileOut = new File(filePath);
            try (FileOutputStream out1 = new FileOutputStream(textFileOut)) {
                out1.write(text.getBytes());
                out1.flush();
            }
        } catch (Exception e) {
            System.out.println(e.getMessage());
            log.error(e);
        }
    }

    void synchronization(double ra, double dec) {
        try {
            Chooser choose = new Chooser();
            String dflt_driver = "ScopeSim";
            String driverID = choose.choose(dflt_driver + ".Telescope");
            if (driverID == null || driverID.length() == 0) {
                driverID = dflt_driver + ".Telescope";
            }
            System.out.println("Driver: " + driverID);
            Thread.sleep(10000L);
            Telescope telescope = new Telescope(driverID);
            System.out.println("setupDialog() ");
            telescope.setupDialog();
            if (!telescope.connected()) {
                System.out.println(telescope.name() + " Not Connected! ");
                System.out.println("setConnected() ");
                telescope.setConnected(true);
            } else {
                System.out.println(telescope.name() + " Connected...");
            }
            Thread.sleep(10000L);
            if (telescope.canSetTracking() && !telescope.tracking()) {
                telescope.setTracking(true);
            }
            boolean prop = telescope.canSync();
            System.out.println("\n Property canSync: " + (prop ? "Yes" : "No"));
            System.out.println("\n syncToCoordinates() ");
            telescope.syncToCoordinates(ra, dec);
            Thread.sleep(10000L);
        } catch (Exception e) {
            System.out.println("synchronization - " + e);
            log.error("synchronization - " + e);
        }
    }

    /**
     * This method converts decimal degrees to hours, minutes and seconds
     *
     * @param degrees - decimal degrees
     * @return String - string containing hours, minutes and seconds
     */
    private String deg2HMS(double degrees) {
        int prec = 1;
        double power = Math.pow(10.0, prec);

        int negative = 0;
        if (degrees < 0) {
            negative = 1;
            degrees = -1.0 * degrees;
        }//if degrees<0
        double hours = degrees / 15.0;
        int iHours = (int) hours;
        double min = hours * 60.0 - ((double) iHours) * 60.0;
        int iMin = (int) min;
        double seconds = min * 60.0 - ((double) iMin) * 60.0;

        seconds = (double) ((int) (seconds * power)) / power;
        if (negative == 1) {
            iHours *= -1.0;
        }
        return iHours + "h:" + iMin + "m:" + seconds + "s";
    }

    void janclient_exit() {
        String text = "ERROR! JANClient is shutdown.\n"
                + "You must input the name of a image file\n"
                + "in  the jpg or png or bmp or gif or tif format,\n"
                + "for example: \n"
                + "java -Djava.library.path=\"./lib\" -jar JANClient.jar IMG4551.jpg \n"
                + "As results in the current or our (see properties) directory You will get:\n"
                + "IMG4551.ANNOTATED_FULL.jpg - with all identified objects\n"
                + "IMG4551.AXY.fits - fits file with pixel positions \n"
                + "IMG4551.RDLS.fits -fits with RA,DEC coordinates\n"
                + "IMG4551.WCS.fits \n"
                + "IMG4551.csv - the file with callibration data for the image\n"
                + "IMG4551.NEW-IMAGE.fits - the image with callibration (aladin file)\n"
                + "IMG4551.annotations.csv - the list of references\n"
                + "second example:\n"
                + "java -Djava.library.path=\"./lib\" -jar JANClient.jar input.tar \n"
                + "in this case You have packed more as one yours images to tar file\n"
                + "the list of callibration is output.csv\n"
                + "In the log directory you will have the stored information about running"
                + "many thanks\n";
        System.out.print(text);
        System.exit(0);
    }
}
